///////////////////////////////////////////////////////////////////////////////////////////
//
//  ========
//  Alcorn							Alcorn Control Protocols
//  ========
//  McBride							C++ Library
//  ========
//  Inc.
//  ========
//
//  The Alcorn Control Serial Class enables hardware and software developers to control 
//  Alcorn McBride Inc. Show Controllers in the Alcorn 8-Bit and Alcorn 9-Bit Protocols.  
//  
//  Both 16-bit and 32-bit versions of this class are included in this library and
//  are easily toggled for compiling by manipulating the #ifdef variable in the header 
//  file.
// 
///////////////////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "serial.h" 

#ifndef WIN16
#include <winbase.h>
#else
#include <conio.h>
#include <dos.h>
#include <limits.h>
#endif

#ifdef _DEBUG
#undef THIS_FILE
static char BASED_CODE THIS_FILE[] = __FILE__;
#endif       

///////////////////////////////////////////////////////////////////////////////////////////
// The Constructor and Destructor for CAlcornControl perform system-dependent functions 
///////////////////////////////////////////////////////////////////////////////////////////

CAlcornControl::CAlcornControl()				// CAlcornControl Constructor
{
	nPort = INVALID ; // assume port no good until initialized
}

CAlcornControl::~CAlcornControl()				// CAlcornControl Destructor
{ 
	Close() ; // make sure interrupt vectors returned before stack pops off instance of class
}

///////////////////////////////////////////////////////////////////////////////////////////
// Open() and Close() simply open and close Comm ports.  COM 1-4 are currently supported.
///////////////////////////////////////////////////////////////////////////////////////////

BOOL CAlcornControl::Open( UINT nNewPort, UINT nNewBaudRate , UINT nNewFormat )	
	{
	char strComPort[4][5] = { "COM1", "COM2", "COM3", "COM4" } ;

	if ( nNewPort > COM4 ) // support only for com1 - com4
		return FALSE ;

#ifndef WIN16
	hCom = CreateFile(strComPort[nNewPort],
		GENERIC_READ | GENERIC_WRITE,
		0,				// comm devices must be opened w/exclusive access
		NULL,			// no security attributes
		OPEN_EXISTING,	// comm devices must use OPEN_EXISTING
		NULL,			// Use Standard COM Port Buffers
		NULL			// hTemplate must be NULL for comm devices
		);
	
		FlushFileBuffers(hCom);

	////////////////////////////////////////////////////////////////////////////////////////////
	// If there is an error opening the COM port, dwError is set equal to the 
	// numerical value of the last error.  This is mainly for debug purposes.
	////////////////////////////////////////////////////////////////////////////////////////////
	
		if (hCom == INVALID_HANDLE_VALUE)	
			{
				dwError = GetLastError();
			}

	////////////////////////////////////////////////////////////////////////////////////////////
	// bSuccess is set equal to the success of the Win32 API GetCommState() function 
	// call.  If there is a problem receiving the state of the specified COM port, bSuccess is
	// set to FALSE, otherwise TRUE.
	////////////////////////////////////////////////////////////////////////////////////////////

		bSuccess = GetCommState(hCom, &dcb);

		if (!bSuccess) 
		{
			// TODO:  Handle the Error
		}
		
#else
		if ( ( nPort = OpenComm( strComPort[nNewPort], UINT_MAX, 128 ) ) < 0 )
			return FALSE ;

		if ( ( ComStatus = GetCommState( nPort, &dcb ) ) < 0 )
			return FALSE ;
#endif

	////////////////////////////////////////////////////////////////////////////////////////////
	// The following code fills in a temporary DCB structure with the new parameters for the
	// COM port.
	////////////////////////////////////////////////////////////////////////////////////////////

		dcb.BaudRate = nNewBaudRate;
		dcb.ByteSize = 8;
		dcb.StopBits = ONESTOPBIT;

		if (nNewFormat == EIGHTH)
		{	
			dcb.Parity = NOPARITY;
			bIs9TH = FALSE;
		}
		else
		{
			dcb.Parity = SPACEPARITY; // SPACE, *** NOT *** MARK
			bIs9TH = TRUE;
		}
		
	////////////////////////////////////////////////////////////////////////////////////////////
	// Open() now tries to set the state of the specified COM port with the API call
	// SetCommState().
	////////////////////////////////////////////////////////////////////////////////////////////

#ifndef WIN16
		bSuccess = SetCommState(hCom, &dcb);

		if (!bSuccess)
		{
				dwError = GetLastError();
		}
#else
		if ( ( ComStatus = SetCommState( &dcb ) ) < 0 ) 
			return FALSE ;
		
	    int nDataBits = 8 ;
		int nStopBits = 1 ;  

		switch( nPort )
			{
			case COM1	:	ioPort = 0x3F8 ; break ;
			case COM2	:	ioPort = 0x2F8 ; break ;
			case COM3	:	ioPort = 0x3E8 ; break ;
			case COM4	:	ioPort = 0x2E8 ; break ;
			}

	    LCRData = ( nDataBits - 5 ) | ( ( nStopBits == 1 ) ? 0 : 4 ) | 0x28 ; 
	    
	    THR = ioPort + 0 ;   
		RBR = ioPort + 0 ;
	    LCR = ioPort + 3 ;
	    LSR = ioPort + 5 ; 

		return TRUE ;
#endif
		
	////////////////////////////////////////////////////////////////////////////////////////////
	// Finally, we do some cleaning-up and set some permanent variables to the values of the
	// COM port.  We also set some of the read timeouts for the port.
	////////////////////////////////////////////////////////////////////////////////////////////

		nBaudRate = nNewBaudRate;
		nFormat = nNewFormat;

#ifndef WIN16
		nPort = nNewPort;
		TimeOuts.ReadIntervalTimeout = TimeOuts.ReadTotalTimeoutMultiplier = MAXDWORD;

		return bSuccess;
#endif
	}			

void CAlcornControl::Close( )
	{
	if ( nPort == INVALID ) // get the heck out if port already closed
		return ;

	////////////////////////////////////////////////////////////////////////////////////////////
	// If there is an error closing the COM port, dwError is set equal to the 
	// numerical value of the last error.  This is mainly for debug purposes.
	////////////////////////////////////////////////////////////////////////////////////////////
#ifndef WIN16
		bSuccess = CloseHandle(hCom);

		if (!bSuccess)
		{
			dwError = GetLastError();
		}
#else
	CloseComm( nPort ) ;
#endif

	nPort = INVALID ; // close port officially with class
	}

////////////////////////////////////////////////////////////////////////////////////////////
// SetSource() and SetTarget() set the Source and Target addresses for 9th bit operation
////////////////////////////////////////////////////////////////////////////////////////////


void CAlcornControl::SetSource( BYTE NewSource )
	{
		Source = NewSource | 0x80 ;	// force top bit high
	}

void CAlcornControl::SetTarget( BYTE NewTarget )
	{
		Target = NewTarget ;
	}


///////////////////////////////////////////////////////////////////////////////////////////
// SendByte() and RecvByte() send and receive single bytes in a particular format
////////////////////////////////////////////////////////////////////////////////////////////


BOOL CAlcornControl::SendMessage( CString sendMessage )
{
	if ( nPort == INVALID ) // get the heck out if port not open
	return FALSE ;

#ifndef WIN16
	int i;

	//////////////////////////////////////////////////////////////////////////////
	// If the port was opened as an 9th bit port, the following segments of code 
	// are used.  First, the Source address is sent with MARKPARITY.  Then, the
	// Target address is sent with SPACEPARITY.   
	//////////////////////////////////////////////////////////////////////////////

	if (bIs9TH)
	{
	
		bSuccess = SendByte( Source , MARKPARITY );

		if (!bSuccess)
		{
			dwError = GetLastError();
		}

		bSuccess = SendByte( Target , SPACEPARITY );
	
		if (!bSuccess)
		{
			dwError = GetLastError();
		}

	}
	
	//////////////////////////////////////////////////////////////////////////////
	// This is the only segment of code that Alcorn 8 Bit uses.  It sends all of 
	// the data bytes in the sendMessage string one at a time.  The Alcorn 9 Bit
	// Control protocol uses this code as well.
	//////////////////////////////////////////////////////////////////////////////


	for(i=0;i<sendMessage.GetLength();i++)
	{
		bSuccess = SendByte(sendMessage[i],SPACEPARITY);

		if (!bSuccess)
		{
			dwError = GetLastError();
		}

		if (bIs9TH)
		{
			Checksum += sendMessage[i];
		}
	}

	//////////////////////////////////////////////////////////////////////////////
	// If the port was opened as an 9th bit port, the following segments of code 
	// are used.  The Checksum is sent   
	//////////////////////////////////////////////////////////////////////////////

	if (bIs9TH)
	{
		bSuccess = SendByte( ( ((Checksum+Source+Target) >> 1) & 0x7F ) , MARKPARITY);

		if (!bSuccess)
		{
			dwError = GetLastError();
		}
	
		Checksum = 0x00;
	}
	
	return bSuccess;
#else
	BYTE checkSum = Source ;

	if (bIs9TH)
		{
	    while ( !( ( LSRData = inp( LSR ) ) & THRE ) || !( LSRData & TEMT ) ) ;
	    
		outp( LCR , LCRData & ~0x10 ) ;     		// 9th bit protocol - set databits, stopbits, and parity 
	    outp( THR, Source ) ;   					// write source address byte to port 
	    
	    while ( !( ( LSRData = inp( LSR ) ) & THRE ) || !( LSRData & TEMT ) ) ;
	    
	    outp( LCR , LCRData | 0x10 ) ;		   		// 8th bit protocol - set databits, stopbits, and parity 
		
	    outp( THR, Target ) ;   					// write target address byte to port 

		while ( !( inp( LSR ) & THRE ) )	;		// wait till first register can take a new byte
		}

	for ( int i = 0 ; i < sendMessage.GetLength() ; i++ )  // send all message bytes in 8th bit protocol 
		{
		checkSum += outp( THR, sendMessage[i] ) ; 
		while ( !( inp( LSR ) & THRE ) )	;		// wait till first register can take a new byte
		}     
		
	if (bIs9TH)
		{
	    while ( !( ( LSRData = inp( LSR ) ) & THRE ) || !( LSRData & TEMT ) ) ;
	    
		outp( LCR , LCRData & ~0x10 ) ;     		// 9th bit protocol - set databits, stopbits, and parity 
		outp( THR, ( checkSum >> 1 ) & 0x7F ) ; 	// write checksum to port
		   
		while ( !( ( LSRData = inp( LSR ) ) & THRE ) || !( LSRData & TEMT ) ) ;
	    
	    outp( LCR , LCRData | 0x10 ) ;		   		// 8th bit protocol - set databits, stopbits, and parity
		}
    
	if ( GetCommError( nPort, &comStat ) < 0 ) // check for errors
		return FALSE ;

	return TRUE ;
#endif
}


BOOL CAlcornControl::SendByte( BYTE sendByte , UINT nParity )
{
	if ( nPort == INVALID ) // get the heck out if port not open
		return FALSE ;

#ifndef WIN16
	DWORD dwNumWritten ;			// Number of bytes written to port

	CString sendMessage=sendByte;	// Unfortunately, we must create a 1 byte CString
									// and then strip it of all data except the byte.
									// Why can't we just send the byte?  Who Knows!!
									// Ask Microsoft....
	
	//////////////////////////////////////////////////////////////////////////////
	// If the port has been opened as a 9TH bit port, the following IF statement
	// sets the parity to the value sent in nParity.
	//////////////////////////////////////////////////////////////////////////////
	
	if (bIs9TH && nParity != dcb.Parity  )  
	{
		
		Sleep(2);

		dcb.Parity = nParity;

		bSuccess = SetCommState(hCom, &dcb);

		if (!bSuccess) 
		{
			dwError = GetLastError();
		}

	}

	//////////////////////////////////////////////////////////////////////////////
	// The byte is now written to the COM Port.
	//////////////////////////////////////////////////////////////////////////////
	
	bSuccess = WriteFile(hCom,(LPCVOID)sendMessage, 1 ,&dwNumWritten,NULL);  // The byte is written to the port and the number of actual bytes 
																				 // written is saved in dwNumWritten
	if (!bSuccess)  // checks for write error
	{
		dwError = GetLastError();
	}
			
	return bSuccess; 
#else
	if (bIs9TH && nParity != dcb.Parity  )  
		{
		dcb.Parity = nParity ; // update parity in class

	    while ( !( ( LSRData = inp( LSR ) ) & THRE ) || !( LSRData & TEMT ) ) ;

		if ( nParity == MARKPARITY )
			outp( LCR , LCRData & ~0x10 ) ;	// 9th bit protocol - set databits, stopbits, and parity 
		else // SPACE PARITY	    
		    outp( LCR , LCRData | 0x10 ) ; 	// 9th bit protocol - set databits, stopbits, and parity 
		}

	outp( THR, sendByte ) ;

	while ( !( inp( LSR ) & THRE ) )	;		// wait till first register can take a new byte

	if ( GetCommError( nPort, &comStat ) < 0 ) // check for errors
		return FALSE ;

	return TRUE ;
#endif	
} 


////////////////////////////////////////////////////////////////////////////////////////////
// SendMessage() and RecvMessage() send and receive messages in a particular format
////////////////////////////////////////////////////////////////////////////////////////////

BOOL CAlcornControl::RecvMessage( BYTE Buffer[] , UINT nMsgLength , UINT nTimeOut )
{

	/////////////////////////////////////////////////////////////////////////
	// This routine simply reads the specified number of bytes from the port
	// and places them in Buffer.  If there is a hardware or timeout failure
	// in reading the byte, the function returns FALSE, else if returns TRUE.
	/////////////////////////////////////////////////////////////////////////

	if ( nPort == INVALID ) // get the heck out if port not open
		return FALSE ;

	for(UINT x=0; x < nMsgLength; x++)
	{
		
		int nResult = RecvByte(nTimeOut);

		if (nResult < 0)
			return FALSE;

		Buffer[x] = (BYTE)nResult;

	}

	return TRUE;

}

int CAlcornControl::RecvByte( UINT nTimeOut )
{
	if ( nPort == INVALID ) // get the heck out if port not open
		return FALSE ;

#ifndef WIN16
	DWORD dwNumRead ;
	LPVOID lpgotByte ;
	unsigned char gotByte ;

	TimeOuts.ReadTotalTimeoutConstant = (DWORD)nTimeOut;
	SetCommTimeouts(hCom , &TimeOuts);
		
	lpgotByte = &gotByte ;



	bSuccess = ReadFile(hCom,lpgotByte, 1,&dwNumRead,NULL);



	if (!bSuccess)
	{
		return COMM_ERROR;
	}
	else if (dwNumRead == 0)
	{
		return TIME_OUT;
	}
	else
	{
		return (int)gotByte;
	}
#else
	union _REGS inRegs ;
	union _REGS outRegs ;
	BYTE nTickCount ;
	WORD nClockTicks = nTimeOut / 55 ;		// 55 milliseconds for every clock tick

	BYTE inBuffer[2] ;

	inRegs.h.ah = 0 ;						// function 0

	_int86( 0x1a, &inRegs, &outRegs ) ;		// get initial count

	nTickCount = outRegs.h.dl ;				// save initial count

	// for duration of timeout

	while ( nClockTicks > 0 ) 					// while ticks are left
		{
		// try to read one byte

		if ( GetCommError( nPort, &comStat ) < 0 ) // check for errors
			return FALSE ;

		int nCharsRead = ReadComm( nPort, (LPSTR)&inBuffer, 1 ) ;

		// if there was an error

		if ( nCharsRead < 0 )
			return COMM_ERROR ; // hardware error

		if ( GetCommError( nPort, &comStat ) < 0 ) // check for errors
			return FALSE ;

		// if we got a byte

		if ( nCharsRead > 0 )
			return (int)inBuffer[0] ;

		_int86( 0x1a, &inRegs, &outRegs ) ;	// get current tick count

		if ( nTickCount != outRegs.h.dl ) 	// if they're not the same
			{								// then one tick has gone by
			nTickCount = outRegs.h.dl ;		// set to the current count

			nClockTicks-- ;					// and diminish one tick
			}
		}

	// never did get a byte, time out!

	return TIME_OUT;
#endif	
}

